package view;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

import javax.swing.*;

import view.gui.Container;

import model.Model;

@SuppressWarnings("serial")
public class GraphicsEngine extends JFrame implements Runnable, MouseMotionListener, MouseListener, KeyListener {
	public static final int WIDTH = 840, HEIGHT = 480; // canvas sizes
	protected static final Color clearcolor = Color.black;

	protected int insl, insr, inst, insb;
	protected BufferedImage canvas;
	protected Graphics2D g2d;
	protected boolean initialized = false; // true when initialization
											// complete...

	private long lasttime;
	private int framedelay = 10; // ms
	protected float delta;

	protected Model model;

	protected Vector<FrameUpdateL> frameupdatelisteners;
	protected Vector<MouseMotionL> mousemotionlisteners;
	protected Vector<MouseL> mouselisteners;
	protected Vector<KeyL> keylisteners;

	public interface MouseL {
		public boolean mouseClicked(int x, int y);
	}

	public interface MouseMotionL {
		public boolean mouseMoved(int x, int y);

		public boolean mouseDragged(int x, int y);

		public boolean mouseDown(int x, int y);

		public boolean mouseUp(int x, int y);
	}

	public interface KeyL {
		public boolean keyTyped(char letter);

		public boolean keyPressed(int code);
		public boolean keyReleased(int code);
	}

	public interface FrameUpdateL {
		public void updateFrame(float delta);
	}

	public double coordScreenToWorldX(int x) { return model.scrollX() + x/Model.zoom; }
	public double coordScreenToWorldY(int y) { return model.scrollY() + y/Model.zoom; }
	
	public void addMouseMotionL(MouseMotionL l) {
		mousemotionlisteners.add(l);
	}

	public void addMouseL(MouseL l) {
		mouselisteners.add(l);
	}

	public void addKeyL(KeyL l) {
		keylisteners.add(l);
	}

	public void addFrameUpdateL(FrameUpdateL l) {
		frameupdatelisteners.add(l);
	}

	void invokeFrameUpdateListeners() {
		for (FrameUpdateL l : frameupdatelisteners) {
			l.updateFrame(delta);
		}
	}

	// blending

	public static Color blendLight(Color col, float light) {
		int r = col.getRed();
		int g = col.getGreen();
		int b = col.getBlue();

		float rat = 1;

		if (light < 0.5f) {
			rat = (0.5f - light) / 0.5f;
			r = r + (int) ((0f - (float) r) * rat);
			g = g + (int) ((0f - (float) g) * rat);
			b = b + (int) ((0f - (float) b) * rat);
		} else if (light > 0.5f) {
			rat = (light - 0.5f) / 0.5f;
			r = r + (int) ((255 - (float) r) * rat);
			g = g + (int) ((255 - (float) g) * rat);
			b = b + (int) ((255 - (float) b) * rat);
		}

		if (r > 255) {
			r = 255;
		}
		if (r < 0) {
			r = 0;
		}
		if (g > 255) {
			g = 255;
		}
		if (g < 0) {
			g = 0;
		}
		if (b > 255) {
			b = 255;
		}
		if (b < 0) {
			b = 0;
		}

		return new Color(r, g, b);
	}

	public static int blendLight(int rgb, float light) {
		return rgb;
	}

	public static Color blend(Color backg, Color foreg) { // foreg's Alpha
		if (foreg == null || foreg.getAlpha() == 0) {
			return backg;
		}
		if (foreg.getAlpha() == 255 || backg == null) {
			return foreg;
		}

		int r = backg.getRed();
		int g = backg.getGreen();
		int b = backg.getBlue();
		float a = (255 - foreg.getAlpha()) / 255f;

		r = (int) (r * a + foreg.getRed() * (1 - a));
		g = (int) (g * a + foreg.getGreen() * (1 - a));
		b = (int) (b * a + foreg.getBlue() * (1 - a));

		// if (backg.getAlpha() != 255)
		// System.out.println("[view.GraphicsEngine blend(..)]true: backg.getAlpha() != 255");
		return new Color(r, g, b, 255);
		// return new Color(r,g,b, backg.getAlpha());
	}

	public static Color blenda(Color backg, Color foreg) { // foreg's Alpha
		if (foreg == null || foreg.getAlpha() == 0) {
			return backg;
		}
		if (foreg.getAlpha() == 255 || backg == null) {
			return foreg;
		}

		int r = backg.getRed();
		int g = backg.getGreen();
		int b = backg.getBlue();
		float a = (255 - foreg.getAlpha()) / 255f;

		r = (int) (r * a + foreg.getRed() * (1 - a));
		g = (int) (g * a + foreg.getGreen() * (1 - a));
		b = (int) (b * a + foreg.getBlue() * (1 - a));

		// if (backg.getAlpha() != 255)
		// System.out.println("[view.GraphicsEngine blend(..)]true: backg.getAlpha() != 255");
		return new Color(r, g, b, backg.getAlpha());
		// return new Color(r,g,b, backg.getAlpha());
	}

	// continue with usual stuff

	protected synchronized void onUpdate() {
	}

	protected synchronized void onRender() {
	} // PLASE put a syncrhonized over canvas because of reRender!

	public GraphicsEngine(Model model) {
		super();
		initialized = false;
		this.model = model;

		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(this);

		frameupdatelisteners = new Vector<FrameUpdateL>();
		mousemotionlisteners = new Vector<MouseMotionL>();
		mouselisteners = new Vector<MouseL>();
		keylisteners = new Vector<KeyL>();

		lasttime = System.currentTimeMillis();
	}

	public void set(int x, int y, Color cl) {
		if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT) {
			return;
		}
		if (cl == null) {
			return;
		}
		canvas.setRGB(x, y, cl.getRGB());
	}

	public void seta(int x, int y, Color cl) {
		if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT) {
			return;
		}
		if (cl == null) {
			return;
		}
		canvas.setRGB(x, y, (blend(get(x, y), cl)).getRGB());
	}

	public void setrgb(int x, int y, int rgb) {
		if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT) {
			return;
		}
		canvas.setRGB(x, y, rgb);
	}

	public Color get(int x, int y) {
		if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT) {
			return clearcolor;
		}
		return new Color(canvas.getRGB(x, y));
	}

	public int getrgb(int x, int y) {
		if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT) {
			return clearcolor.getRGB();
		}
		return canvas.getRGB(x, y);
	}

	public Model getModel() {
		return model;
	}

	public void reRender() {// onRender(); TODO
	}

	public void paint(Graphics g) {
		if (!initialized) {
			System.out.println("GE/repaint: Not yet initialzed!");
			return;
		}
		if (canvas == null) {
			System.out.println("GE/repaint: Canvas is null!");
			return;
		}

		/*
		 * g.setColor(new Color(0,0,0, 10)); g.fillRect(insl, insr, WIDTH+insl,
		 * HEIGHT+inst );
		 */

		/*
		 * for (int x=0; x<WIDTH; x++) { for (int y=0; y<HEIGHT; y++) {
		 * canvas.setRGB(x, y, canvas.getRGB(x, y)); } }
		 */
		g.drawImage(canvas, insl, inst, WIDTH + insl, HEIGHT + inst, 0, 0,
				WIDTH, HEIGHT, null);
	}

	public void run() {
		while (true) {
			long millis = System.currentTimeMillis();
			float diff = millis - lasttime;

			if (diff > framedelay) {

				delta = diff / 1000f; // ms -> s
				onUpdate();

				g2d.setColor(clearcolor);
				// System.out.println("GE.run.canvas.width: "+canvas.getWidth()
				// + "| .height: " + canvas.getHeight());
				g2d.fillRect(0, 0, canvas.getWidth(), canvas.getHeight());
				onRender();
				repaint();

				lasttime = System.currentTimeMillis();
			} else {
				long sleeptime = (long) (framedelay - diff);

				try {
					Thread.currentThread().sleep(sleeptime);
				} catch (Exception ex) {
					ex.printStackTrace();
					System.out.println("Couldnt sleep!");
				}
			}
		}

		/*
		 * while (initialized) { float diff = System.currentTimeMillis() -
		 * lasttime; if (diff > framedelay) { delta = diff/1000f; onUpdate();
		 * 
		 * g2d.setColor(clearcolor); g2d.fillRect(0, 0, WIDTH, HEIGHT);
		 * onRender(); repaint();
		 * 
		 * lasttime = System.currentTimeMillis(); } else { long sleeptime =
		 * (long) (framedelay - diff - 1); try { Thread.sleep(sleeptime); }
		 * catch (Exception ex) { System.out.println("Too much coffee today!");
		 * } } }
		 */
	}

	public void start() {
		canvas = new BufferedImage(WIDTH, HEIGHT, BufferedImage.TYPE_4BYTE_ABGR);
		g2d = (Graphics2D) canvas.getGraphics();

		JFrame hateframe = new JFrame();

		hateframe.pack();
		insl = hateframe.getInsets().left;
		insr = hateframe.getInsets().right;
		inst = hateframe.getInsets().top;
		insb = hateframe.getInsets().bottom;
		hateframe = null;

		setTitle(".");
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setSize(WIDTH + insl + insr, HEIGHT + inst + insb);
		setVisible(true);

		addMouseMotionListener(this);

		initialized = true;
		lasttime = System.currentTimeMillis();
		new Thread(this).start();
	}

	// event spam
	public synchronized final void mouseClicked(MouseEvent ev) {
		for (MouseL l : mouselisteners) {
			l.mouseClicked(ev.getX() - insl, ev.getY() - inst);
		}
	}

	public synchronized final void keyTyped(KeyEvent ev) {
		if (!initialized) {
			return;
		}
		for (KeyL l : keylisteners) {
			l.keyTyped(ev.getKeyChar());
		}
	}

	public synchronized final void keyPressed(KeyEvent ev) {
		if (!initialized) {
			return;
		}
		for (KeyL l : keylisteners) {
			l.keyPressed(ev.getKeyCode());
		}
	}

	public synchronized final void mouseDragged(MouseEvent ev) {
		if (!initialized) {
			return;
		}
		for (MouseMotionL l : mousemotionlisteners) {
			l.mouseDragged(ev.getX() - insl, ev.getY() - inst);
		}
	}

	public synchronized final void mouseMoved(MouseEvent ev) {
		if (!initialized) {
			return;
		}
		for (MouseMotionL l : mousemotionlisteners) {
			l.mouseMoved(ev.getX() - insl, ev.getY() - inst);
		}
	}

	public synchronized final void keyReleased(KeyEvent ev) {
		if (!initialized) {
			return;
		}
		for (KeyL l : keylisteners) {
			l.keyReleased(ev.getKeyCode());
		}
	}

	public synchronized final void mouseEntered(MouseEvent ev) {
	}

	public synchronized final void mouseExited(MouseEvent ev) {
	}

	public synchronized final void mousePressed(MouseEvent ev) {
		if (!initialized) {
			return;
		}
		for (MouseMotionL l : mousemotionlisteners) {
			l.mouseDown(ev.getX() - insl, ev.getY() - inst);
		}
	}

	public synchronized final void mouseReleased(MouseEvent ev) {
		if (!initialized) {
			return;
		}
		for (MouseMotionL l : mousemotionlisteners) {
			l.mouseUp(ev.getX() - insl, ev.getY() - inst);
		}
	}

	// font spam
	private final static byte[][] FAIL = { { 1, 1, 1, 1, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 1, 0, 1, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 1, 0, 1, 1 }, { 1, 0, 0, 0, 1 }, { 1, 1, 1, 1, 1 } };
	private final static byte[][] A = { { 0, 0, 1, 0, 0 }, { 0, 1, 0, 1, 0 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 1, 1, 1, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 } };
	private final static byte[][] B = { { 1, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 1, 1, 1, 0 },
			{ 1, 0, 0, 0, 1 }, { 1, 1, 1, 1, 0 } };
	private final static byte[][] C = { { 0, 0, 1, 1, 0 }, { 0, 1, 0, 0, 1 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
			{ 0, 1, 0, 0, 1 }, { 0, 0, 1, 1, 0 } };
	private final static byte[][] D = { { 1, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 1, 1, 1, 0 } };
	private final static byte[][] E = { { 1, 1, 1, 1, 1 }, { 1, 0, 0, 0, 0 },
			{ 1, 1, 1, 1, 0 }, { 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 1, 1, 1, 1 } };
	private final static byte[][] F = { { 1, 1, 1, 1, 1 }, { 1, 0, 0, 0, 0 },
			{ 1, 1, 1, 1, 0 }, { 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 } };
	private final static byte[][] G = { { 0, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 }, { 1, 0, 1, 1, 0 },
			{ 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] H = { { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 1, 1, 1, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 } };
	private final static byte[][] I = { { 0, 1, 1, 1, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] J = { { 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 },
			{ 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] K = { { 1, 0, 0, 0, 1 }, { 1, 0, 0, 1, 0 },
			{ 1, 0, 1, 0, 0 }, { 1, 1, 0, 0, 0 }, { 1, 0, 1, 0, 0 },
			{ 1, 0, 0, 1, 0 }, { 1, 0, 0, 1, 0 } };
	private final static byte[][] L = { { 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 1, 1, 1, 1 } };
	private final static byte[][] M = { { 1, 0, 0, 0, 1 }, { 1, 1, 0, 1, 1 },
			{ 1, 0, 1, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 } };
	private final static byte[][] N = { { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 1, 0, 0, 1 }, { 1, 0, 1, 0, 1 }, { 1, 0, 0, 1, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 } };
	private final static byte[][] O = { { 0, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] P = { { 1, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 1, 1, 1, 0 }, { 1, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 } };
	private final static byte[][] Q = { { 0, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 1, 0 }, { 0, 1, 1, 0, 1 } };
	private final static byte[][] R = { { 1, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 1, 1, 1, 0 }, { 1, 0, 1, 0, 0 },
			{ 1, 0, 0, 1, 0 }, { 1, 0, 0, 1, 0 } };
	private final static byte[][] S = { { 0, 1, 1, 1, 1 }, { 1, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 }, { 0, 1, 1, 1, 0 },
			{ 0, 0, 0, 0, 1 }, { 1, 1, 1, 1, 0 } };
	private final static byte[][] T = { { 1, 1, 1, 1, 1 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 } };
	private final static byte[][] U = { { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] V = { { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 0, 1, 0, 1, 0 }, { 0, 0, 1, 0, 0 } };
	private final static byte[][] W = { { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 1, 0, 1 }, { 0, 1, 0, 1, 0 } };
	private final static byte[][] X = { { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 0, 1, 0, 1, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 1, 0, 1, 0 }, { 1, 0, 0, 0, 1 } };
	private final static byte[][] Y = { { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 0, 1, 0, 1, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 } };
	private final static byte[][] Z = { { 1, 1, 1, 1, 1 }, { 0, 0, 0, 0, 1 },
			{ 0, 0, 0, 1, 0 }, { 0, 0, 1, 0, 0 }, { 0, 1, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 1, 1, 1, 1 } };
	private final static byte[][] SPACE = { { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 } };
	private final static byte[][] EXCL = { { 0, 1, 0, 0, 0 },
			{ 0, 1, 0, 0, 0 }, { 0, 1, 0, 0, 0 }, { 0, 1, 0, 0, 0 },
			{ 0, 1, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0 } };
	private final static byte[][] QUEST = { { 0, 0, 0, 0, 0 },
			{ 0, 1, 1, 0, 0 }, { 1, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 1, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0 } };
	private final static byte[][] COLON = { { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0 }, { 0, 0, 0, 0, 0 } };
	private final static byte[][] NUM0 = { { 0, 1, 1, 1, 0 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 0, 1, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] NUM1 = { { 0, 0, 0, 1, 0 },
			{ 0, 0, 1, 1, 0 }, { 0, 1, 0, 1, 0 }, { 0, 0, 0, 1, 0 },
			{ 0, 0, 0, 1, 0 }, { 0, 0, 0, 1, 0 }, { 0, 0, 0, 1, 0 } };
	private final static byte[][] NUM2 = { { 0, 1, 1, 1, 0 },
			{ 1, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 }, { 0, 0, 0, 1, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 1, 0, 0, 0 }, { 1, 1, 1, 1, 1 } };
	private final static byte[][] NUM3 = { { 1, 1, 1, 1, 1 },
			{ 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 },
			{ 0, 1, 1, 1, 1 }, { 0, 0, 0, 0, 1 }, { 1, 1, 1, 1, 0 } };
	private final static byte[][] NUM4 = { { 1, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 }, { 1, 0, 1, 0, 0 },
			{ 1, 0, 1, 0, 0 }, { 1, 1, 1, 1, 1 }, { 0, 0, 1, 0, 0 } };
	private final static byte[][] NUM5 = { { 1, 1, 1, 1, 1 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 }, { 0, 1, 1, 1, 0 },
			{ 0, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] NUM6 = { { 0, 1, 1, 1, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
			{ 1, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] NUM7 = { { 1, 1, 1, 1, 1 },
			{ 1, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 }, { 0, 0, 1, 1, 1 },
			{ 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 } };
	private final static byte[][] NUM8 = { { 0, 1, 1, 1, 0 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 }, { 1, 0, 0, 0, 1 },
			{ 0, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] NUM9 = { { 0, 1, 1, 1, 0 },
			{ 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 1 }, { 0, 0, 0, 0, 1 },
			{ 0, 0, 0, 0, 1 }, { 0, 0, 0, 0, 1 }, { 0, 0, 0, 1, 0 } };
	private final static byte[][] DOT = { { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 0, 0, 0 } };
	private final static byte[][] COMMA = { { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0 }, { 1, 0, 0, 0, 0 } };
	private final static byte[][] BRACEIN = { { 0, 0, 0, 1, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 0, 1, 0 } };
	private final static byte[][] BRACEOUT = { { 0, 1, 0, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 1, 0, 0, 0 } };
	private final static byte[][] AT = { { 0, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 },
			{ 1, 0, 0, 0, 1 }, { 1, 0, 1, 0, 1 }, { 1, 0, 1, 1, 1 },
			{ 1, 0, 0, 0, 0 }, { 0, 1, 1, 1, 0 } };
	private final static byte[][] MINUS = { { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 1, 1, 1, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 } };
	private final static byte[][] PLUS = { { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 1, 1, 1, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 0, 0, 0 } };
	private final static byte[][] UNDERSCORE = { { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 1, 1, 1, 1, 1 } };
	private final static byte[][] SEP = { { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 } };
	private final static byte[][] DOLLAR = { { 0, 0, 1, 0, 0 },
			{ 0, 1, 1, 1, 1 }, { 1, 0, 1, 0, 0 }, { 0, 1, 1, 1, 0 },
			{ 0, 0, 1, 0, 1 }, { 1, 1, 1, 1, 0 }, { 0, 0, 1, 0, 0 } };
	private final static byte[][] HASH = { { 0, 0, 0, 0, 0 },
			{ 0, 1, 0, 1, 0 }, { 0, 1, 0, 1, 0 }, { 1, 1, 1, 1, 1 },
			{ 0, 1, 0, 1, 0 }, { 1, 1, 1, 1, 1 }, { 0, 1, 0, 1, 0 } };
	private final static byte[][] SEMICOLON = { { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0 }, { 1, 0, 0, 0, 0 } };
	private final static byte[][] TRIIN = // <
	{ { 0, 0, 0, 0, 0 }, { 0, 0, 0, 1, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 1, 0, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 0, 1, 0 },
			{ 0, 0, 0, 0, 0 } };
	private final static byte[][] TRIOUT = // >
	{ { 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 0, 1, 0 }, { 0, 0, 1, 0, 0 }, { 0, 1, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 } };
	private final static byte[][] ACCENT = // '
	{ { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 } };
	private final static byte[][] SLASH = // /
	{ { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 1 }, { 0, 0, 0, 1, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 1, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 } };
	private final static byte[][] BACKSLASH = // \
	{ { 0, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 }, { 0, 1, 0, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 0, 1, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 } };
	private final static byte[][] PERCENT = // %
	{ { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 1, 0, 0, 1 },
			{ 0, 0, 0, 1, 0 }, { 0, 0, 1, 0, 0 }, { 0, 1, 0, 0, 0 },
			{ 1, 0, 0, 1, 0 } };
	private final static byte[][] STRING = // "
	{ { 0, 1, 0, 1, 0 }, { 0, 1, 0, 1, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 } };
	private final static byte[][] POW = // ^
	{ { 0, 0, 1, 0, 0 }, { 0, 1, 0, 1, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 } };
	private final static byte[][] STAR = // *
	{ { 0, 1, 0, 1, 0 }, { 0, 0, 1, 0, 0 }, { 0, 1, 0, 1, 0 },
			{ 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 } };
	private final static byte[][] EQU = { { 0, 0, 0, 0, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 0, 0, 0, 0 }, { 1, 1, 1, 1, 0 }, { 0, 0, 0, 0, 0 },
			{ 1, 1, 1, 1, 0 }, { 0, 0, 0, 0, 0 } };
	private final static byte[][] SQBRIN = // square brace in [
	{ { 0, 0, 1, 1, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 1, 0 } };
	private final static byte[][] SQBROUT = // ]
	{ { 0, 1, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 1, 1, 0, 0 } };
	private final static byte[][] WABRIN = // {
	{ { 0, 0, 1, 1, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 1, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 1, 0 } };
	private final static byte[][] WABROUT = // }
	{ { 0, 1, 1, 0, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 0, 1, 1, 0 }, { 0, 0, 1, 0, 0 }, { 0, 0, 1, 0, 0 },
			{ 0, 1, 1, 0, 0 } };

	public final static byte[][] HAPPY = { { 0, 0, 0, 0, 0 },
			{ 0, 1, 0, 1, 0 }, { 0, 1, 0, 1, 0 }, { 0, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 1 }, { 0, 1, 1, 1, 0 }, { 0, 0, 0, 0, 0 } };
	public final static byte[][] UNHAPPY = { { 0, 0, 0, 0, 0 },
			{ 0, 1, 0, 1, 0 }, { 0, 1, 0, 1, 0 }, { 0, 0, 0, 0, 0 },
			{ 0, 1, 1, 1, 0 }, { 1, 0, 0, 0, 1 }, { 0, 0, 0, 0, 0 } };
	public final static byte[][] DVER = { { 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 },
			{ 1, 0, 0, 0, 0 }, { 1, 0, 0, 0, 0 } };

	public static byte[][] charToMap(char x) {
		switch (x) {
		case 'a':
		case 'A':
			return A;

		case 'b':
		case 'B':
			return B;

		case 'c':
		case 'C':
			return C;

		case 'd':
		case 'D':
			return D;

		case 'e':
		case 'E':
			return E;

		case 'f':
		case 'F':
			return F;

		case 'g':
		case 'G':
			return G;

		case 'h':
		case 'H':
			return H;

		case 'i':
		case 'I':
			return I;

		case 'j':
		case 'J':
			return J;

		case 'k':
		case 'K':
			return K;

		case 'l':
		case 'L':
			return L;

		case 'm':
		case 'M':
			return M;

		case 'n':
		case 'N':
			return N;

		case 'o':
		case 'O':
			return O;

		case 'p':
		case 'P':
			return P;

		case 'q':
		case 'Q':
			return Q;

		case 'r':
		case 'R':
			return R;

		case 's':
		case 'S':
			return S;

		case 't':
		case 'T':
			return T;

		case 'u':
		case 'U':
			return U;

		case 'v':
		case 'V':
			return V;

		case 'w':
		case 'W':
			return W;

		case 'x':
		case 'X':
			return X;

		case 'y':
		case 'Y':
			return Y;

		case 'z':
		case 'Z':
			return Z;

		case '0':
			return NUM0;

		case '1':
			return NUM1;

		case '2':
			return NUM2;

		case '3':
			return NUM3;

		case '4':
			return NUM4;

		case '5':
			return NUM5;

		case '6':
			return NUM6;

		case '7':
			return NUM7;

		case '8':
			return NUM8;

		case '9':
			return NUM9;

		case '.':
			return DOT;

		case '!':
			return EXCL;

		case '?':
			return QUEST;

		case ' ':
			return SPACE;

		case ':':
			return COLON;

		case ';':
			return SEMICOLON;

		case ',':
			return COMMA;

		case '@':
			return AT;

		case '-':
			return MINUS;

		case '+':
			return PLUS;

		case '_':
			return UNDERSCORE;

		case '|':
			return SEP;

		case '$':
			return DOLLAR;

		case '#':
			return HASH;

		case '\'':
			return ACCENT;

		case '/':
			return SLASH;

		case '\\':
			return BACKSLASH;

		case '%':
			return PERCENT;

		case '"':
			return STRING;

		case '^':
			return POW;

		case '*':
			return STAR;

		case '=':
			return EQU;

		case '(':
			return BRACEIN;

		case ')':
			return BRACEOUT;

		case '<':
			return TRIIN;

		case '>':
			return TRIOUT;

		case '[':
			return SQBRIN;

		case ']':
			return SQBROUT;

		case '{':
			return WABRIN;

		case '}':
			return WABROUT;

			// my own special chars
		case 128:
			return HAPPY;

		case 129:
			return UNHAPPY;

		case 130:
			return DVER;
		}
		return FAIL;
	}

	public int drawChar(int x, int y, byte[][] map, Color cl) {
		return drawChar(x, y, map, cl, 2);
	}
	public int drawChar(int x, int y, byte[][] map, Color cl, int s) {

		for (int my = 0; my < map.length; my++) {
			for (int mx = 0; mx < map[0].length; mx++) {
				if (map[my][mx] == 1) {
					for (int cx = 0; cx < s; cx++) {
						for (int cy = 0; cy < s; cy++) {
							set(cx + mx * s + x, cy + my * s + y, cl);
						}
					}

					/*
					 * set(x+mx*s , y+my*s , cl); set(x+mx*s , y+my*s+1, cl);
					 * set(x+mx*s+1, y+my*s , cl); set(x+mx*s+1, y+my*s+1, cl);
					 */
				}
			}
		}
		return map[0].length * s;
	}
	public int drawChar(int x, int y, char chr, Color cl, int s) {
		return drawChar(x, y, charToMap(chr), cl, s);
	}
	public static int gsdrawChar(int x, int y, byte[][] map, Color cl, int s, GSnip onwhat) {
		for (int my = 0; my < map.length; my++) { // vv
			for (int mx = 0; mx < map[0].length; mx++) { // iterate thought all
															// chars

				if (map[my][mx] == 1) {
					for (int cx = 0; cx < s; cx++) {
						for (int cy = 0; cy < s; cy++) {
							onwhat.set(cx + mx * s + x, cy + my * s + y, cl);

							// onwhat.set(mx+x+cx, my+y+cy, cl);
						}
					}
				}

			}
		}
		return map[0].length * s;
	}
	public static int gsdrawChar(int x, int y, char chr, Color cl, int s,
			GSnip onwhat) {
		return gsdrawChar(x, y, charToMap(chr), cl, s, onwhat);
	}
	public static int textlen(String text, int scale) {
		return text.length() * (5 * (scale + 2));
	}
	public void text(int x, int y, String text) {
		text(x, y, text, 2);
	}
	public void text(int x, int y, String text, int scale) {
		text(x, y, text, Color.LIGHT_GRAY, scale);
	}
	public void text(int x, int y, String text, Color cl) {
		text(x, y, text, cl, 2);
	}
	public void text(int x, int y, String text, Color cl, int scale) {
		int xoff = 0;

		for (int i = 0; i < text.length(); i++) {
			xoff += scale
					+ drawChar(x + xoff, y, charToMap(text.charAt(i)), cl,
							scale);
		}
	}
	public static int gstext(int x, int y, String text, Color cl, int scale, GSnip onwhat) {
		int xoff = 0;

		for (int i = 0; i < text.length(); i++) {
			xoff += scale
					+ gsdrawChar(x + xoff, y, charToMap(text.charAt(i)), cl,
							scale, onwhat);
		}
		return xoff;
	}

	
	
	
}
